@@ -100,7 +100,6 @@ fn assert_json_eq(results: json::Json, mut expected: json::Json, message: String
     }
 }
 
-
 fn run_raw_json_tests<F: Fn(Json, Json) -> ()>(json_data: &str, run: F) {
     let items = match Json::from_str(json_data) {
         Ok(Json::Array(items)) => items,
@@ -319,14 +318,7 @@ fn color3_hsl() {
 /// color3_keywords.json is different: R, G and B are in 0..255 rather than 0..1
 #[test]
 fn color3_keywords() {
-    run_color_tests(include_str!("css-parsing-tests/color3_keywords.json"), |c| {
-        match c {
-            Ok(Color::RGBA(RGBA { red: r, green: g, blue: b, alpha: a }))
-            => [r * 255., g * 255., b * 255., a].to_json(),
-            Ok(Color::CurrentColor) => "currentColor".to_json(),
-            Err(()) => Json::Null,
-        }
-    });
+    run_color_tests(include_str!("css-parsing-tests/color3_keywords.json"), |c| c.ok().to_json())
 }
 
 
@@ -387,25 +379,28 @@ fn serializer(preserve_comments: bool) {
     });
 }
 
-
 #[test]
 fn serialize_current_color() {
     let c = Color::CurrentColor;
     assert!(c.to_css_string() == "currentColor");
 }
 
-
 #[test]
 fn serialize_rgb_full_alpha() {
-    let c = Color::RGBA(RGBA { red: 1.0, green: 0.9, blue: 0.8, alpha: 1.0 });
-    assert!(c.to_css_string() == "rgb(255, 230, 204)");
+    let c = Color::RGBA(RGBA::new(255, 230, 204, 255));
+    assert_eq!(c.to_css_string(), "rgb(255, 230, 204)");
 }
 
-
 #[test]
 fn serialize_rgba() {
-    let c = Color::RGBA(RGBA { red: 0.1, green: 0.2, blue: 0.3, alpha: 0.5 });
-    assert!(c.to_css_string() == "rgba(26, 51, 77, 0.5)");
+    let c = Color::RGBA(RGBA::new(26, 51, 77, 32));
+    assert_eq!(c.to_css_string(), "rgba(26, 51, 77, 0.125)");
+}
+
+#[test]
+fn serialize_rgba_two_digit_float_if_roundtrips() {
+    let c = Color::RGBA(RGBA::from_floats(0., 0., 0., 0.5));
+    assert_eq!(c.to_css_string(), "rgba(0, 0, 0, 0.5)");
 }
 
 #[test]
@@ -574,8 +569,8 @@ fn identifier_serialization() {
 impl ToJson for Color {
     fn to_json(&self) -> json::Json {
         match *self {
-            Color::RGBA(RGBA { red, green, blue, alpha }) => {
-                [red, green, blue, alpha].to_json()
+            Color::RGBA(ref rgba) => {
+                [rgba.red, rgba.green, rgba.blue, rgba.alpha].to_json()
             },
             Color::CurrentColor => "currentColor".to_json(),
         }
